#include <mdarrays.h>
#include <internal.h>

// Note: Much of the code in this file is not used.  The package was originally much more
//       ambitious, but Valgrind essentially made a lot of it redundant.


// See the description of global variables in mdarrays.h (before the extern declarations)
const uint ARRAY_LAST_INDEX_MARKER = UINT_MAX-1;

const uint _mdarrays_access_type_real = 0;
const uint _mdarrays_access_type_long_double = 1;
const uint _mdarrays_access_type_double = 2;
const uint _mdarrays_access_type_float = 3;
const uint _mdarrays_access_type_long_int = 5;
const uint _mdarrays_access_type_int = 6;
const uint _mdarrays_access_type_short_int = 7;
const uint _mdarrays_access_type_unsigned_long_int = 9;
const uint _mdarrays_access_type_unsigned_int = 10;
const uint _mdarrays_access_type_unsigned_short_int = 11;
const uint _mdarrays_access_type_char = 13;
const uint _mdarrays_access_type_unsigned_char = 16;
const uint _mdarrays_access_type_pointer = 18;

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
mdarray_info_struct *
_mdarrays_get_mdarray_info(char *cfile, uint cline, void *array, const char *array_name)
{
char *char_pointer = NULL;
mdarray_info_struct *mdarray_info;

APD_PUSH_FUNCTION(4, "array @ %lX", (unsigned long int)array);
MDARRAYS_ASSERT(array != NULL);
MDARRAYS_ASSERT(array_name != NULL);

char_pointer = (char *)array;
char_pointer = char_pointer - sizeof(mdarray_info_struct);
mdarray_info = (mdarray_info_struct *)char_pointer;
APD_REPORT(5, "name=%s", mdarray_info->name);
MDARRAYS_ASSERT(mdarray_info->tag == ARRAY_INFO_STRUCT_TAG,
        "(array %s @%lX) this is not a valid mdarray (tag=%d).  Did you allocate this array?",
        array_name, (unsigned long int)array, mdarray_info->tag);

APD_POP_FUNCTION();
return mdarray_info;
}

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Check bounds (internal version, called by _mdarrays_assert_bounds())
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _check_bounds_internal(char *cfile, uint cline, void *array, 
                            mdarray_info_struct *mdarray_info, 
                            const char *array_name, uint line_number, 
                            uint access_index[MDARRAYS_MAX_DIMENSIONS])
{
uint current_dimension;
uint current_index;
uint n_dimensions;

APD_PUSH_FUNCTION(4);

// Read sizes and decend into mdarray
current_dimension = 0;
current_index = access_index[current_dimension];
while ( current_index != ARRAY_LAST_INDEX_MARKER )
  {
    MDARRAYS_ASSERT(current_dimension < MDARRAYS_MAX_DIMENSIONS);
    MDARRAYS_ASSERT(current_index < mdarray_info->sizes[current_dimension]
                    || current_index == ARRAY_LAST_INDEX_MARKER,
                    "line %d: array \"%s\" at %lX accessed with index #%d=%d, "
                    "which is greater than limit of %d",
                    line_number, array_name, (unsigned long int)array, current_dimension,
                    current_index, mdarray_info->sizes[current_dimension]-1);
    current_dimension++;
    current_index = access_index[current_dimension];
  }
n_dimensions = current_dimension;

// Check number of dimensions
APD_REPORT(5, "number of dimensions=%d", n_dimensions);
MDARRAYS_ASSERT(n_dimensions <= mdarray_info->dimensions, "(%d == %d)",
                n_dimensions, mdarray_info->dimensions);

if ( n_dimensions < mdarray_info->dimensions )
  {
    APD_REPORT(3, "Array \"%s\": too few dimensions", array_name);
  }
APD_POP_FUNCTION();
} // end _check_bounds_internal()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Given a variable parameter list with the array indicess,
// return a access_index array
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _generate_access_index(char *cfile, uint cline, 
                            va_list index_list, uint access_index[MDARRAYS_MAX_DIMENSIONS])
{
uint current_dimension, current_index;

APD_PUSH_FUNCTION(4);

// Read sizes and decend into mdarray
current_dimension = 0;
current_index = va_arg(index_list, int);
while ( current_index != ARRAY_LAST_INDEX_MARKER )
  {
    access_index[current_dimension] = current_index;
    current_dimension++;
    current_index = va_arg(index_list, int);
  }
access_index[current_dimension] = ARRAY_LAST_INDEX_MARKER;
APD_POP_FUNCTION();
} // end _generate_access_index()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Check bounds
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_assert_bounds(char *cfile, uint cline, void *array, const char *array_name, ...)
{
va_list index_list;
mdarray_info_struct *mdarray_info;
uint access_index[MDARRAYS_MAX_DIMENSIONS];

APD_PUSH_FUNCTION(4);

// Get info for this array
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, array, array_name);
APD_REPORT(5, "from \"%s\", line %d: examining array \"%s\"", cfile, cline, mdarray_info->name);

va_start(index_list, array_name);
_generate_access_index(cfile, cline, index_list, access_index);
_check_bounds_internal(cfile, cline, array, mdarray_info, array_name, cline, access_index);
va_end(index_list);
APD_POP_FUNCTION();
} // end _mdarrays_assert_bounds()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Get a pointer to an element in a mdarray
// +++++++++++++++++++++++++++++++++++++++++++++++++
void * _mdarrays_get_access_pointer(char *cfile, uint cline, void *array, 
                                    mdarray_info_struct *mdarray_info, 
                                    char *array_name, uint access_index[MDARRAYS_MAX_DIMENSIONS])
{
uint current_dimension;
char *char_pointer;
char **char_pointer2;

APD_PUSH_FUNCTION(4);

// Decend into pointer to return single pointer
char_pointer2 = (char **)array;

for (current_dimension = 0; current_dimension < mdarray_info->dimensions-1; current_dimension++)
  {
    APD_REPORT(5, "pointing to dimension=%d", current_dimension);
    APD_REPORT(5, "char_pointer2 @%lX", (unsigned long int)char_pointer2);
    char_pointer2 = char_pointer2 + access_index[current_dimension];
    APD_REPORT(5, "char_pointer2 @%lX", (unsigned long int)char_pointer2);
    char_pointer2 = (char **)(*char_pointer2);
    APD_REPORT(5, "char_pointer2 @%lX", (unsigned long int)char_pointer2);
  }
// Last dimension
char_pointer = (char *)char_pointer2;
APD_REPORT(5, "char_pointer @%lX", (unsigned long int)char_pointer);
char_pointer = char_pointer + mdarray_info->type_size * access_index[current_dimension];
APD_REPORT(5, "mdarray_info->type_size=%d, access_indexs=%d",
           mdarray_info->type_size, access_index[current_dimension]);
APD_REPORT(5, "char_pointer @%lX", (unsigned long int)char_pointer);

APD_POP_FUNCTION();
return (void *)char_pointer;
} // end _mdarrays_get_access_pointer()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
void * _mdarrays_access(char *cfile, uint cline, void *array, char *array_name, ...)
{
va_list index_list;
uint access_index[MDARRAYS_MAX_DIMENSIONS];  // never used, but needed by _check_bounds_internal
mdarray_info_struct *mdarray_info;
void *access_pointer;

APD_PUSH_FUNCTION(4, "from \"%s\", line %d: mdarray \"%s\" @ %lX", 
                  cfile, cline, array_name, (unsigned long int)array);

// Get info for this array
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, array, array_name);

va_start(index_list, array_name);
_generate_access_index(cfile, cline, index_list, access_index);
_check_bounds_internal(cfile, cline, array, mdarray_info, array_name, cline, access_index);
access_pointer = _mdarrays_get_access_pointer(cfile, cline, array, mdarray_info, array_name, 
                                              access_index);

APD_POP_FUNCTION();
return access_pointer;
} // end _mdarrays_access()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Assign a value to a pointer (internal version)
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _assign_pointer(char *cfile, uint cline, void *array,
                     mdarray_info_struct *mdarray_info,
                     char *array_name, uint access_index[MDARRAYS_MAX_DIMENSIONS],
                     void *value_pointer)
{
void *access_pointer;
char *write_me;
char *read_me;
uint i;

APD_PUSH_FUNCTION(4);
access_pointer = _mdarrays_get_access_pointer(cfile, cline, array, mdarray_info, 
                                              array_name, access_index);

// Copy the right number of bytes over from value_pointer to access_pointer
// NOTE: This is inefficient... could be sped up quite a bit
read_me = (char *)value_pointer;
write_me = (char *)access_pointer;
for (i = 0; i < mdarray_info->type_size; i++)
  {
    write_me[i] = read_me[i];
  }

APD_POP_FUNCTION();
} // end _assign_pointer()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Assign the value pointed to by value_pointer into array
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_assign(char *cfile, uint cline, void *array, 
                      char *array_name, void *value_pointer, ...)
{
va_list index_list;
uint access_index[MDARRAYS_MAX_DIMENSIONS];  // never used, but needed by _check_bounds_internal
mdarray_info_struct *mdarray_info;

APD_PUSH_FUNCTION(4, "from \"%s\", line %d:  mdarray \"%s\" @ %lX", cfile, cline, array_name, (unsigned long int)array);

// Get info for this array
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, array, array_name);

va_start(index_list, value_pointer);
_generate_access_index(cfile, cline, index_list, access_index);
_check_bounds_internal(cfile, cline, array, mdarray_info, array_name, cline, access_index);
_assign_pointer(cfile, cline, array, mdarray_info, array_name, access_index, value_pointer);

APD_POP_FUNCTION();
} // end _mdarrays_assign()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Get a pointer to an element in a nonmdarray
// +++++++++++++++++++++++++++++++++++++++++++++++++
/*
void * _mdarrays_get_nonmdarray_access_pointer(char *cfile, uint cline, void *array, int type, int type_size, int dimensions, int *sizes, int *element)
{
int current_dimension;
char *char_pointer;
char **char_pointer2;

APD_PUSH_FUNCTION(4);

//add code here that returns a pointer to an element given array specifications

APD_POP_FUNCTION();
return (void *)char_pointer;
} // end _mdarrays_get_nonmdarray_access_pointer()
*/
