#include <mdarrays.h>
#include <internal.h>

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Modeled after the glib foreach function.
// Operate on all the members of an array.
// Call mdarrays_foreach_function with each array element
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_foreach_recurse(char *cfile, uint cline, 
                               void *array, uint array_dimensions, uint *array_sizes, 
                               mdarrays_foreach_function foreach_function,
                               void *user_data, uint access_index[MDARRAYS_MAX_DIMENSIONS], 
                               uint dimension)
{
uint i;
APD_PUSH_FUNCTION(5);

// If last dimension, iterate here
if (dimension == array_dimensions - 1)
  {
    MDARRAYS_ASSERT(array_sizes[dimension] > 0);
    for (i = 0; i < array_sizes[dimension]; i++)
      {
        access_index[dimension] = i;
        foreach_function(array, access_index, user_data);
      }
  }
else
  {
    MDARRAYS_ASSERT(array_sizes[dimension] > 0);
    for (i = 0; i < array_sizes[dimension]; i++)
      {
        access_index[dimension] = i;
        _mdarrays_foreach_recurse(cfile, cline, array, array_dimensions, array_sizes, 
                                 foreach_function, user_data, access_index, dimension + 1);
      }
  }
APD_POP_FUNCTION();
} // end _mdarrays_foreach_recurse()

// ----------------------------------------------------------
void _mdarrays_foreach(char *cfile, uint cline, void *array,
                       mdarrays_foreach_function foreach_function, void *user_data)
{
uint access_index[MDARRAYS_MAX_DIMENSIONS];
mdarray_info_struct *mdarray_info;

APD_PUSH_FUNCTION(4, "array @ %lX", (unsigned long int)array);
MDARRAYS_ASSERT(array != NULL);

// Get info for this array
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, array, "unknown");
MDARRAYS_ASSERT(mdarray_info->dimensions > 0);

_mdarrays_foreach_recurse(cfile, cline, array, mdarray_info->dimensions, mdarray_info->sizes, 
                          foreach_function, user_data, access_index, 0);
APD_POP_FUNCTION();
}

// ----------------------------------------------------------
bool _mdarrays_compare_values(char *cfile, uint cline, void *value1, void *value2, uint type)
//<0> _mdarrays_compare_values(cfile, cline, value1, value2, type)
//<0>
//<0> Return true if the value pointed to by value1 equals the value
//<0> pointed to by value2.
{
bool equal = false;
APD_PUSH_FUNCTION(5, "value1 @ %lX, value2 @ %lX", 
                     (unsigned long int)value1, (unsigned long int)value2);
MDARRAYS_ASSERT(value1 != NULL);
MDARRAYS_ASSERT(value2 != NULL);

char char1, char2;
uchar uchar1, uchar2;
bool bool1, bool2;
int int1, int2;
uint uint1, uint2;
long int lint1, lint2;
float float1, float2;
double double1, double2;
long double ldbl1, ldbl2;
real real1, real2;

switch (type)
  {
    case MDARRAYS_CHAR_TYPE:
      char1 = *((char *)value1);
      char2 = *((char *)value2);
      APD_REPORT(5, "Test %d == %d", char1, char2);
      if (! char1 == char2)
        {
          equal = false;
        }
    break;
    case MDARRAYS_UCHAR_TYPE:
      uchar1 = *((unsigned char *)value1);
      uchar2 = *((unsigned char *)value2);
      APD_REPORT(5, "Test %d == %d", uchar1, uchar2);
      if (! uchar1 == uchar2)
        {
          equal = false;
        }
    break;
    case MDARRAYS_BOOL_TYPE:
      bool1 = *((bool *)value1);
      bool2 = *((bool *)value2);
      APD_REPORT(5, "Test %d == %d", bool1, bool2);
      if (! bool1 == bool2)
        {
          equal = false;
        }
    break;
    case MDARRAYS_INT_TYPE:
      int1 = *((int *)value1);
      int2 = *((int *)value2);
      APD_REPORT(5, "Test %d == %d", int1, int2);
      if (! int1 == int2)
        {
          equal = false;
        }
    break;
    case MDARRAYS_UINT_TYPE:
      uint1 = *((unsigned int *)value1);
      uint2 = *((unsigned int *)value2);
      APD_REPORT(5, "Test %ud == %ud", uint1, uint2);
      if (! uint1 == uint2)
        {
          equal = false;
        }
    break;
    case MDARRAYS_LONG_INT_TYPE:
      lint1 = *((long int *)value1);
      lint2 = *((long int *)value2);
      APD_REPORT(5, "Test %ld == %ld", lint1, lint2);
      if (! lint1 == lint2)
        {
          equal = false;
        }
    break;
    case MDARRAYS_FLOAT_TYPE:
      float1 = *((float *)value1);
      float2 = *((float *)value2);
      APD_REPORT(5, "Test %g == %g", float1, float2);
      if (! dm_flt_equals((float)float1, (float)float2))
        {
          equal = false;
        }
    break;
    case MDARRAYS_DOUBLE_TYPE:
      double1 = *((double *)value1);
      double2 = *((double *)value2);
      APD_REPORT(5, "Test %g == %g", double1, double2);
      if (! dm_dbl_equals(double1, double2))
        {
          equal = false;
        }
    break;
    case MDARRAYS_LONG_DOUBLE_TYPE:
      ldbl1 = *((long double *)value1);
      ldbl2 = *((long double *)value2);
      APD_REPORT(5, "Test %Lg == %Lg", ldbl1, ldbl2);
      if (! dm_ldbl_equals(ldbl1, ldbl2))
        {
          equal = false;
        }
    break;
    case MDARRAYS_REAL_TYPE:
      real1 = *((real *)value1);
      real2 = *((real *)value2);
      APD_REPORT(5, "Test %g == %g", real1, real2);
      if (! dm_real_equals(real1, real2))
        {
          equal = false;
        }
    break;

    default:
      MDARRAYS_ASSERT(1==0, "array type %d not supported for _mdarrays_equal", type);
      break;
  }
APD_POP_FUNCTION();
return equal;
} // end _mdarrays_compare_values()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
bool _mdarrays_equal_recurse(char *cfile, uint cline, 
                             void *array1, mdarray_info_struct *array1_info, 
                             void *array2, mdarray_info_struct *array2_info,
                             uint array_dimensions, uint dimension,
                             uint *array_sizes, uint *access_index)
{
uint i;
bool equal = true;
APD_PUSH_FUNCTION(5);

// If last dimension, iterate here
if (dimension == array_dimensions - 1)
  {
    MDARRAYS_ASSERT(array_sizes[dimension] > 0);
    for (i = 0; i < array_sizes[dimension]; i++)
      {
        access_index[dimension] = i;
        void *p1 =
          _mdarrays_get_access_pointer(cfile, cline, array1, array1_info, array1_info->name,
                                       access_index);
        void *p2 =
          _mdarrays_get_access_pointer(cfile, cline, array2, array2_info, array2_info->name,
                                       access_index);
        APD_REPORT(5, "Comparing pointer @ %lX and %lX", 
                   (unsigned long int)p1, (unsigned long int)p2);

        equal = _mdarrays_compare_values(cfile, cline, p1, p2, array1_info->type);
        if (! equal) goto _mdarrays_equal_recurse_end;
      }
  }
else
  {
    MDARRAYS_ASSERT(array_sizes[dimension] > 0);
    for (i = 0; i < array_sizes[dimension]; i++)
      {
        access_index[dimension] = i;
        _mdarrays_equal_recurse(cfile, cline, array1, array1_info, array2, array2_info,
                                array_dimensions, dimension + 1, array_sizes, access_index);
      }
  }

_mdarrays_equal_recurse_end:;
APD_POP_FUNCTION();
return equal;
} // end _mdarrays_equal_recurse()

// ----------------------------------------------------------
bool _mdarrays_equal(char *cfile, uint cline, void *array1, void *array2)
//<0> _mdarrays_equal(cfile, cline, array1, array2)
//<0>
//<0> Check that array1 has the same dimensions as array2, then compare
//<0> every element in the two arrays.  If every element is equal,
//<0> return true, otherwise return false.
{
uint access_index[MDARRAYS_MAX_DIMENSIONS];
mdarray_info_struct *array1_info;
mdarray_info_struct *array2_info;

APD_PUSH_FUNCTION(4, "array1 @ %lX, array2 @ %lX", 
                     (unsigned long int)array1, (unsigned long int)array2);
MDARRAYS_ASSERT(array1 != NULL);
MDARRAYS_ASSERT(array2 != NULL);

// Get info for these arrays
array1_info = _mdarrays_get_mdarray_info(cfile, cline, array1, "unknown");
array2_info = _mdarrays_get_mdarray_info(cfile, cline, array2, "unknown");
MDARRAYS_ASSERT(array1_info->dimensions > 0);
MDARRAYS_ASSERT(array2_info->dimensions > 0);

bool equal = 
_mdarrays_equal_recurse(cfile, cline, array1, array1_info,
                        array2, array2_info, array1_info->dimensions, 0/*dimension*/,
                        array1_info->sizes, access_index);
APD_POP_FUNCTION();
return equal;
}

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
bool _mdarrays_is_mdarray(char *cfile, uint cline, void *array)
{
mdarray_info_struct *mdarray_info;
int return_me;

APD_PUSH_FUNCTION(4, "array @ %lX", (unsigned long int)array);
MDARRAYS_ASSERT(array != NULL);

// Get info for this array
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, array, "unknown");
APD_REPORT(5, "examining array \"%s\"", mdarray_info->name);

return_me = (mdarray_info->tag == ARRAY_INFO_STRUCT_TAG);
APD_REPORT(5, "returning %d", return_me);
APD_POP_FUNCTION();
return return_me;
}

// ==============================================================
// ==============================================================
void _mdarrays_get_type_from_string(char *cfile, uint cline, char *type_string, 
                                    uint *type)
//<0> _mdarrays_get_type_from_string(cfile, cline, type_string
//<0>
//<0> Given type_string, return the type in *type.
{
APD_PUSH_FUNCTION(4);
MDARRAYS_ASSERT(type_string != NULL);
MDARRAYS_ASSERT(type != NULL);
APD_REPORT(5, "type_string=%s", type_string);

// Set type based on type_string
if (strcmp(type_string, "char") == 0)               *type = MDARRAYS_CHAR_TYPE;
else if (strcmp(type_string, "uchar") == 0)         *type = MDARRAYS_UCHAR_TYPE;
else if (strcmp(type_string, "bool") == 0)          *type = MDARRAYS_BOOL_TYPE;
else if (strcmp(type_string, "_Bool") == 0)         *type = MDARRAYS_BOOL_TYPE;
else if (strcmp(type_string, "int") == 0)           *type = MDARRAYS_INT_TYPE;
else if (strcmp(type_string, "uint") == 0)          *type = MDARRAYS_UINT_TYPE;
else if (strcmp(type_string, "unsigned int") == 0)  *type = MDARRAYS_UINT_TYPE;
else if (strcmp(type_string, "long int") == 0)      *type = MDARRAYS_LONG_INT_TYPE;
else if (strcmp(type_string, "float") == 0)         *type = MDARRAYS_FLOAT_TYPE;
else if (strcmp(type_string, "double") == 0)        *type = MDARRAYS_DOUBLE_TYPE;
else if (strcmp(type_string, "long double") == 0)   *type = MDARRAYS_LONG_DOUBLE_TYPE;
else if (strcmp(type_string, "real") == 0)          *type = MDARRAYS_REAL_TYPE;
else
  {
    _mdarrays_abort(cfile, cline, 
                    "expected a basic MDARRAYS type, not %s "
                    "(use OBJECT version of allocation routines?", type_string);
  }

APD_POP_FUNCTION();
}

// ==============================================================
// default abort function
// ==============================================================
void _mdarrays_abort_function(char *file_i, uint line_i, const char *function_i, 
                              char *file, uint line, char *fmt, ...)
{
va_list args;
va_start(args, fmt);
fprintf(MDARRAYS_ERROR_STREAM, "%s:%d: ", file, line);
vfprintf(MDARRAYS_ERROR_STREAM, fmt, args);
fprintf(MDARRAYS_ERROR_STREAM,
        "\n(mdarrays library at:\n%s:%d:%s when error detected)\n", file_i, line_i, function_i);
fprintf(MDARRAYS_ERROR_STREAM, "\n");
va_end(args);
fprintf(MDARRAYS_ERROR_STREAM, "Intentionally causing segfault for debugger to catch\n");
fflush(MDARRAYS_ERROR_STREAM);
*((uint *)(0)) = 0;  // trigger a Valgrind stack trace if Valgrind is running

fprintf(MDARRAYS_ERROR_STREAM, "Segfault failed, exiting.\n");
exit(EXIT_FAILURE);
}

// ==========================================
// ==========================================
void _mdarrays_copy_1d(char *cfile, uint cline, void *from, uint from_offset,
                            void *to, uint to_offset)
//<0> _mdarrays_copy_1d(cfile, cline, from, from_offset, to, to_offset)
//<0>
//<0> Copy elements of from array to to array starting at element from_offset
//<0> in the from array and to_offset in the to array.  
//<0> Stop if end of either array is reached.
{
char *from_char = (char *)from;
uchar *from_uchar = (uchar *)from;
bool *from_bool = (bool *)from;
int *from_int = (int *)from;
uint *from_uint = (uint *)from;
long int *from_long_int = (long int *)from;
float *from_float = (float *)from;
double *from_double = (double *)from;
long double *from_long_double = (long double *)from;
real *from_real = (real *)from;
//gpointer_l *from_gpointer = (gpointer_l *)from;

char *to_char = (char *)to;
uchar *to_uchar = (uchar *)to;
bool *to_bool = (bool *)to;
int *to_int = (int *)to;
uint *to_uint = (uint *)to;
long int *to_long_int = (long int *)to;
float *to_float = (float *)to;
double *to_double = (double *)to;
long double *to_long_double = (long double *)to;
real *to_real = (real *)to;
//gpointer_l *to_gpointer = (gpointer_l *)to;

APD_PUSH_FUNCTION(5, "from @ %lX, to @ %lX, from_offset=%d, to_offset=%d", 
                  (unsigned long int)from, (unsigned long int)to, from_offset, to_offset);
MDARRAYS_ASSERT(from != NULL);
MDARRAYS_ASSERT(to != NULL);

//<1> Call _mdarrays_get_mdarray_info() to get info for arrays
mdarray_info_struct *from_info = _mdarrays_get_mdarray_info(cfile, cline, from, "unknown");
mdarray_info_struct *to_info   = _mdarrays_get_mdarray_info(cfile, cline, to  , "unknown");

//<1> Check that offsets are within bounds
MDARRAYS_ASSERT(from_offset < from_info->sizes[0]);
MDARRAYS_ASSERT(to_offset   < to_info->sizes[0]);

//<1> Check that from and to have the same type
MDARRAYS_ASSERT(from_info->type == to_info->type,
                "MDARRAYS copy requires that arrays have the same type");

//<1> Set from_index to from_offset and to_index to to_offset
uint from_index = from_offset;
uint to_index = to_offset;

APD_REPORT(5, "from @ %lX, to @ %lX, from_offset=%d, to_offset=%d", 
           (unsigned long int)from, (unsigned long int)to, from_offset, to_offset);

//<1> While not at the end of either array
while (true)
  {
    //<2> Copy current element from from array to to array

    APD_REPORT(5, "copy1d: from_index=%d/%d, to_index=%d/%d", 
               from_index, from_info->sizes[0], to_index, to_info->sizes[0]);
    switch (from_info->type)
      {
        case MDARRAYS_CHAR_TYPE:          
          to_char[to_index] = from_char[from_index];
          break;
        case MDARRAYS_UCHAR_TYPE:          
          to_uchar[to_index] = from_uchar[from_index];
          break;
        case MDARRAYS_BOOL_TYPE:          
          to_bool[to_index] = from_bool[from_index];
          break;
        case MDARRAYS_INT_TYPE:           
          to_int[to_index] = from_int[from_index];
          break;
        case MDARRAYS_UINT_TYPE:          
          to_uint[to_index] = from_uint[from_index];
          break;
        case MDARRAYS_LONG_INT_TYPE:      
          to_long_int[to_index] = from_long_int[from_index];
          break;
        case MDARRAYS_FLOAT_TYPE:         
          to_float[to_index] = from_float[from_index];
          break;
        case MDARRAYS_DOUBLE_TYPE:        
          to_double[to_index] = from_double[from_index];
          break;
        case MDARRAYS_LONG_DOUBLE_TYPE:   
          to_long_double[to_index] = from_long_double[from_index];
          break;
        case MDARRAYS_REAL_TYPE:          
          to_real[to_index] = from_real[from_index];
          break;
        case MDARRAYS_GPOINTER_TYPE:      
          _mdarrays_abort(cfile, cline, "gpointer type not supported for copy yet", 
                          from_info->type);
          // How to copy gpointer depends... do you want to copy by
          // value or by reference?
          //
          //to_gpointer[to_index] = from_gpointer[from_index];
          break;
        default:
          _mdarrays_abort(cfile, cline, "unknown type: %d", from_info->type);
      }

    //<2> Increment from and to indices
    from_index++;
    to_index++;

    if (from_index == from_info->sizes[0]) break;
    if (to_index   == to_info->sizes[0]) break;
  }
APD_POP_FUNCTION();
}

