#include <mdarrays.h>
#include <internal.h>

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Print an element from a 1D array
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_print_1D_element(char *cfile, uint cline, FILE *file, void *array, 
                                uint type, uint type_size, uint element, 
                                print_style_struct *print_style, bool delimiter_on)
{
APD_PUSH_FUNCTION(5, "printing 1D pointer array @ %lX", (unsigned long int)array);
MDARRAYS_ASSERT(file != NULL);
MDARRAYS_ASSERT(array != NULL);
uint print_type = print_style->type;
uint print_width = print_style->width;
uint print_precision = print_style->precision;
char delimiter = print_style->delimiter;

// Check print parameters consistency
switch(type)
  {
    case MDARRAYS_CHAR_TYPE:
    case MDARRAYS_UCHAR_TYPE:
    case MDARRAYS_BOOL_TYPE:
        MDARRAYS_ASSERT(print_type == MDARRAYS_PRINTF_TYPE_c
                          || print_type == MDARRAYS_PRINTF_TYPE_d
                          || print_type == MDARRAYS_PRINTF_TYPE_x
                          || print_type == MDARRAYS_PRINTF_TYPE_X
                          || print_type == MDARRAYS_PRINTF_TYPE_lX
                          || print_type == MDARRAYS_PRINTF_TYPE_u,
                          "incompatible printf type specifier (%d) for this type (%d)",
                          print_type, type);
        break;
    case MDARRAYS_INT_TYPE:
    case MDARRAYS_UINT_TYPE:
        MDARRAYS_ASSERT(print_type == MDARRAYS_PRINTF_TYPE_c
                          || print_type == MDARRAYS_PRINTF_TYPE_d
                          || print_type == MDARRAYS_PRINTF_TYPE_x
                          || print_type == MDARRAYS_PRINTF_TYPE_X
                          || print_type == MDARRAYS_PRINTF_TYPE_lX
                          || print_type == MDARRAYS_PRINTF_TYPE_u,
                          "incompatible printf type specifier (%d) for this type (%d)",
                          print_type, type);
        break;
    case MDARRAYS_LONG_INT_TYPE:
        MDARRAYS_ASSERT(print_type == MDARRAYS_PRINTF_TYPE_c
                          || print_type == MDARRAYS_PRINTF_TYPE_d
                          || print_type == MDARRAYS_PRINTF_TYPE_ld
                          || print_type == MDARRAYS_PRINTF_TYPE_x
                          || print_type == MDARRAYS_PRINTF_TYPE_X
                          || print_type == MDARRAYS_PRINTF_TYPE_lX
                          || print_type == MDARRAYS_PRINTF_TYPE_u,
                          "incompatible printf type specifier (%d) for this type (%d)",
                          print_type, type);
        break;
    case MDARRAYS_FLOAT_TYPE:
        MDARRAYS_ASSERT(print_type == MDARRAYS_PRINTF_TYPE_e
                          || print_type == MDARRAYS_PRINTF_TYPE_E
                          || print_type == MDARRAYS_PRINTF_TYPE_f
                          || print_type == MDARRAYS_PRINTF_TYPE_g
                          || print_type == MDARRAYS_PRINTF_TYPE_G,
                          "incompatible printf type specifier (%d) for this type (%d)",
                          print_type, type);
        break;
    case MDARRAYS_DOUBLE_TYPE:
        MDARRAYS_ASSERT(print_type == MDARRAYS_PRINTF_TYPE_e
                          || print_type == MDARRAYS_PRINTF_TYPE_E
                          || print_type == MDARRAYS_PRINTF_TYPE_f
                          || print_type == MDARRAYS_PRINTF_TYPE_g
                          || print_type == MDARRAYS_PRINTF_TYPE_G,
                          "incompatible printf type specifier (%d) for this type (%d)",
                          print_type, type);
        break;
    case MDARRAYS_LONG_DOUBLE_TYPE:
        MDARRAYS_ASSERT(print_type == MDARRAYS_PRINTF_TYPE_Le
                          || print_type == MDARRAYS_PRINTF_TYPE_LE
                          || print_type == MDARRAYS_PRINTF_TYPE_Lf
                          || print_type == MDARRAYS_PRINTF_TYPE_Lg
                          || print_type == MDARRAYS_PRINTF_TYPE_LG,
                          "incompatible printf type specifier (%d) for this type (%d)",
                          print_type, type);
        break;
    case MDARRAYS_REAL_TYPE:
        MDARRAYS_ASSERT(print_type == MDARRAYS_PRINTF_TYPE_e
                          || print_type == MDARRAYS_PRINTF_TYPE_E
                          || print_type == MDARRAYS_PRINTF_TYPE_f
                          || print_type == MDARRAYS_PRINTF_TYPE_g
                          || print_type == MDARRAYS_PRINTF_TYPE_G
                          || print_type == MDARRAYS_PRINTF_TYPE_Le
                          || print_type == MDARRAYS_PRINTF_TYPE_LE
                          || print_type == MDARRAYS_PRINTF_TYPE_Lf
                          || print_type == MDARRAYS_PRINTF_TYPE_Lg
                          || print_type == MDARRAYS_PRINTF_TYPE_LG,
                          "incompatible printf type specifier (%d) for this type (%d)",
                          print_type, type);
        break;
    case MDARRAYS_GPOINTER_TYPE:
        MDARRAYS_ASSERT(print_type == MDARRAYS_PRINTF_TYPE_lX,
                          "incompatible printf type specifier (%d) for this type (%d)",
                          print_type, type);
    break;
    default:
      APD_DIE("unknown type: %d", type);
    break;
  } // end switch(type)

// Extract the value of the element into a long double or long int
long double flt_value = 0.0L;
long int int_value = 0;
if (   type == MDARRAYS_CHAR_TYPE
    || type == MDARRAYS_UCHAR_TYPE
    || type == MDARRAYS_BOOL_TYPE
    || type == MDARRAYS_INT_TYPE
    || type == MDARRAYS_UINT_TYPE
    || type == MDARRAYS_LONG_INT_TYPE
    || type == MDARRAYS_GPOINTER_TYPE)
  { /* Integer types */
    switch(type)
      {
        case MDARRAYS_CHAR_TYPE:
            int_value = (long int)(((char *)array)[element]);
            break;
        case MDARRAYS_UCHAR_TYPE:
            int_value = (long int)(((uchar *)array)[element]);
            break;
        case MDARRAYS_BOOL_TYPE:
            int_value = (long int)(((bool *)array)[element]);
            break;
        case MDARRAYS_INT_TYPE:
            int_value = (long int)(((int *)array)[element]);
            break;
        case MDARRAYS_UINT_TYPE:
            int_value = (long int)(((uint *)array)[element]);
            break;
        case MDARRAYS_LONG_INT_TYPE:
            int_value = (long int)(((long int *)array)[element]);
            break;
        case MDARRAYS_GPOINTER_TYPE:
            int_value = (long int)((unsigned long int)array + 
                                  (unsigned long int)(type_size * element));
            break;
        default:
            _mdarrays_abort(cfile, cline, "unknown type");
      }
    switch(print_type)
      {
        case MDARRAYS_PRINTF_TYPE_c:
            fprintf(file, "%c", (char)(int_value));
            APD_REPORT(5, "array[%d] == %c", element, (char)(int_value));
            break;
#define print_case_int(print_type, cast) \
        case MDARRAYS_PRINTF_TYPE_ ## print_type: \
            fprintf(file, "%*.*" #print_type " ", print_width, print_precision, \
                          (cast)(int_value)); \
            APD_REPORT(5, "array[%d] == (%d.%d%s) %*.*" #print_type, element, print_width, \
                          print_precision, #print_type, print_width, print_precision, \
                          (cast)(int_value)); \
            break;
        print_case_int(d, int)
        print_case_int(ld, long int)
        print_case_int(x, unsigned int)
        print_case_int(X, unsigned int)
        print_case_int(lX, unsigned long int)
        print_case_int(u, unsigned int)

        default:
            _mdarrays_abort(cfile, cline, "unknown integer printf type");
      }
  }
else /* floating types */
  {
    switch(type)
      {
        case MDARRAYS_FLOAT_TYPE:
            flt_value = (long double)(((float *)array)[element]);
            break;
        case MDARRAYS_DOUBLE_TYPE:
            flt_value = (long double)(((double *)array)[element]);
            break;
        case MDARRAYS_LONG_DOUBLE_TYPE:
            flt_value = (long double)(((long double *)array)[element]);
            break;
        case MDARRAYS_REAL_TYPE:
            flt_value = (long double)(((real *)array)[element]);
            break;

        default:
            _mdarrays_abort(cfile, cline, "unknown floating type");
      } // end switch(type)
    switch(print_type)
      {
#define print_case_flt(print_type, cast) \
        case MDARRAYS_PRINTF_TYPE_ ## print_type: \
            fprintf(file, "%*.*" #print_type, print_width, print_precision, \
                          (cast)(flt_value)); \
            APD_REPORT(5, "array[%d] == %*.*" #print_type, element, print_width, \
                          print_precision, (cast)(flt_value)); \
            break;
        print_case_flt(e, double)
        print_case_flt(Le, long double)
        print_case_flt(E, double)
        print_case_flt(LE, long double)
        print_case_flt(f, double)
        print_case_flt(Lf, long double)
        print_case_flt(g, double)
        print_case_flt(Lg, long double)
        print_case_flt(G, double)
        print_case_flt(LG, long double)

        default:
            _mdarrays_abort(cfile, cline, "unknown floating printf type");
      }
  }

// Print the delimiter if needed
if (delimiter_on)
  {
    fprintf(file, "%c", delimiter);
  }
APD_POP_FUNCTION();
} // end _mdarrays_print_1D_element()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Print a 1D array
// 1D arrays are all the same
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_print_1D_array(char *cfile, uint cline, FILE *file, void *array,
                              uint type, uint type_size, uint array_size,
                              print_style_struct *print_style)
{
uint col, N_cols;

APD_PUSH_FUNCTION(4, "printing 1D pointer array @ %lX", (unsigned long int)array);
MDARRAYS_ASSERT(array != NULL);
N_cols = array_size;
MDARRAYS_ASSERT(N_cols > 0);

for (col = 0; col < N_cols; col++)
  {
    bool delimiter_on = !(col == N_cols - 1);
    _mdarrays_print_1D_element(cfile, cline, file, array, type, type_size, 
                               col, print_style, delimiter_on);
  } // end for col

APD_POP_FUNCTION();
} // end _mdarrays_print_1D_array()


// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Print a 1D array with identifier comment tags
// tags is an array of strings, one for each array element
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_print_1D_array_tagged(char *cfile, uint cline, FILE *file, void *array, uint type, 
                                     uint type_size, uint array_size, char **tags,
                                     print_style_struct *print_style)
{
uint col, N_cols;

APD_PUSH_FUNCTION(4, "printing 1D pointer array @ %lX", (unsigned long int)array);
MDARRAYS_ASSERT(array != NULL);
MDARRAYS_ASSERT(tags != NULL);
N_cols = array_size;
MDARRAYS_ASSERT(N_cols > 0);

for (col = 0; col < N_cols; col++)
  {
    char *tag = tags[col];
    MDARRAYS_ASSERT(tag != NULL);
    bool delimiter_on = !(col == N_cols - 1);
    _mdarrays_print_1D_element(cfile, cline, file, array, type, type_size, 
                               col, print_style, delimiter_on);
    fprintf(file, " # %s\n", tag);
  } // end for col

APD_POP_FUNCTION();
} // end _mdarrays_print_1D_array_tagged()


// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Print a 2D pointer array (but not a mdarray, because they may
// passed pieces of a larger mdarray from _mdarrays_print_3D
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_print_2D_pointer_array(char *cfile, uint cline, FILE *file, void *array, uint type, 
                                      uint type_size, uint dimensions, uint *sizes,
                                      print_style_struct *print_style)
{
int row, N_rows;

APD_PUSH_FUNCTION(4, "printing 2D pointer array @ %lX", (unsigned long int)array);

MDARRAYS_ASSERT(array != NULL);

N_rows = sizes[0];
MDARRAYS_ASSERT(N_rows > 0);

for (row = 0; row < N_rows; row++)
  {
    _mdarrays_print_1D_array(cfile, cline, file, (void *)(((double **)array)[row]),
                             type, type_size, sizes[1], print_style);
    fprintf(file, "\n");
  }

APD_POP_FUNCTION();
} // end _mdarrays_print_2D_pointer_array()


// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_print_recurse(char *cfile, uint cline, FILE *file, void *array, 
                             uint type, uint type_size, 
                             uint original_dimensions, uint dimensions, 
                             uint *sizes, uint *current_index,
                             print_style_struct *print_style)
{
uint i;

APD_PUSH_FUNCTION(4, "(recursion) printing array @ %lX", (unsigned long int)array);
MDARRAYS_ASSERT(array != NULL);

if (dimensions > 2)
  {
    // Iterate through elements of this dimension and recurse
    MDARRAYS_ASSERT(sizes[0] > 0);
    for (i = 0; i < sizes[0]; i++)
      {
        current_index[original_dimensions - dimensions] = i;
        _mdarrays_print_recurse(cfile, cline, file, (void *)(((double **)array)[i]), 
                                type, type_size, original_dimensions, dimensions - 1, 
                                sizes + 1, current_index, print_style);
      } // end for i
  }
else
  {
    MDARRAYS_ASSERT(dimensions == 2);
    // Print the table label that describes the higher element indices
    for (i = 0; i < original_dimensions - 2; i++)
      {
        fprintf(file, "[%d]", current_index[i]);
      }
    fprintf(file, "%c", print_style->line_delimiter);
    _mdarrays_print_2D_pointer_array(cfile, cline, file, array, type, type_size, 
                                     dimensions, sizes, print_style);
  }

APD_POP_FUNCTION();
} // end _mdarrays_print_recurse()


// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Print an mdarray of any dimension
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_print(char *cfile, uint cline, FILE *file, void *array)
{
mdarray_info_struct *mdarray_info;
uint current_index[MDARRAYS_MAX_DIMENSIONS]; // Used all the way down the call stack to remember the current element
APD_PUSH_FUNCTION(4, "printing array @ %lX", (unsigned long int)array);
MDARRAYS_ASSERT(array != NULL);

// Get info for this array
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, (void *)array, "unknown");
APD_REPORT(5, "examining array \"%s\"", mdarray_info->name);
MDARRAYS_ASSERT(mdarray_info->dimensions > 0);
if (mdarray_info->dimensions >= 2)
  {
    _mdarrays_print_recurse(cfile, cline,
                            file, array, mdarray_info->type, mdarray_info->type_size,
                            mdarray_info->dimensions, mdarray_info->dimensions, 
                            mdarray_info->sizes, current_index, &(mdarray_info->print_style));
  }
else
  {
    _mdarrays_print_1D_array(cfile, cline,
                             file, array, mdarray_info->type, mdarray_info->type_size, 
                             mdarray_info->sizes[mdarray_info->dimensions - 1],
                             &(mdarray_info->print_style));
  }

APD_POP_FUNCTION();
}

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// Print an mdarray with identifier tags for each element
// (currently works for 1-D only)
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_print_tagged(char *cfile, uint cline, FILE *file, void *array, char **tags)
{
mdarray_info_struct *mdarray_info;
// Used all the way down the call stack to remember the current element:
//int current_index[MDARRAYS_MAX_DIMENSIONS]; 
APD_PUSH_FUNCTION(4, "printing array @ %lX", (unsigned long int)array);
MDARRAYS_ASSERT(file != NULL);
MDARRAYS_ASSERT(array != NULL);
MDARRAYS_ASSERT(tags != NULL);

// Get info for this array
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, (void *)array, "unknown");
MDARRAYS_ASSERT(mdarray_info->dimensions == 1, 
                  "this function currently only works for 1D arrays");
APD_REPORT(5, "examining array \"%s\"", mdarray_info->name);
MDARRAYS_ASSERT(mdarray_info->dimensions > 0);
if (mdarray_info->dimensions >= 2)
  {
  }
else
  {
    _mdarrays_print_1D_array_tagged(cfile, cline,
                                    file, array, mdarray_info->type, mdarray_info->type_size,
                                    mdarray_info->sizes[mdarray_info->dimensions - 1], tags,
                                    &(mdarray_info->print_style));
  }

APD_POP_FUNCTION();
} // end _mdarrays_print_tagged()


